@@ -2,7 +2,7 @@
  * This file is part of OpenTTD.
  * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
  * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
- * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
+ * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <https://www.gnu.org/licenses/old-licenses/gpl-2.0>.
  */
 
 /** @file squirrel.cpp the implementation of the Squirrel class. It handles all Squirrel-stuff and gives a nice API back to work with. */
@@ -10,6 +10,7 @@
 #include "../stdafx.h"
 #include "../debug.h"
 #include "squirrel_std.hpp"
+#include "../error_func.h"
 #include "../fileio_func.h"
 #include "../string_func.h"
 #include "script_fatalerror.hpp"
@@ -17,15 +18,10 @@
 #include <sqstdaux.h>
 #include <../squirrel/sqpcheader.h>
 #include <../squirrel/sqvm.h>
-#include "../core/alloc_func.hpp"
+#include "../core/math_func.hpp"
+#include "../core/string_consumer.hpp"
 
-/**
- * In the memory allocator for Squirrel we want to directly use malloc/realloc, so when the OS
- * does not have enough memory the game does not go into unrecoverable error mode and kill the
- * whole game, but rather let the AI die though then we need to circumvent MallocT/ReallocT.
- *
- * So no #include "../safeguards.h" here as is required, but after the allocator's implementation.
- */
+#include "../safeguards.h"
 
 /*
  * If changing the call paths into the scripting engine, define this symbol to enable full debugging of allocations.
@@ -34,7 +30,9 @@
  */
 
 struct ScriptAllocator {
-	size_t allocated_size;   ///< Sum of allocated data size
+private:
+	std::allocator<uint8_t> allocator;
+	size_t allocated_size = 0; ///< Sum of allocated data size
 	size_t allocation_limit; ///< Maximum this allocator may use before allocations fail
 	/**
 	 * Whether the error has already been thrown, so to not throw secondary errors in
@@ -42,7 +40,7 @@ struct ScriptAllocator {
 	 * throw a Squirrel error so the Squirrel stack can be dumped, however that gets
 	 * allocated by this allocator and then you might end up in an infinite loop.
 	 */
-	bool error_thrown;
+	bool error_thrown = false;
 
 	static const size_t SAFE_LIMIT = 0x8000000; ///< 128 MiB, a safe choice for almost any situation
 
@@ -50,42 +48,51 @@ struct ScriptAllocator {
 	std::map<void *, size_t> allocations;
 #endif
 
-	void CheckLimit() const
+	/**
+	 * Checks whether an allocation is allowed by the memory limit set for the script.
+	 * @param requested_size The requested size that was requested to be allocated.
+	 * @throws Script_FatalError When memory may not be allocated (limit reached, except for error handling).
+	 */
+	void CheckAllocationAllowed(size_t requested_size)
 	{
-		if (this->allocated_size > this->allocation_limit) throw Script_FatalError("Maximum memory allocation exceeded");
+		/* When an error has been thrown, we are allocating just a bit of memory for the stack trace. */
+		if (this->error_thrown) return;
+
+		if (this->allocated_size + requested_size <= this->allocation_limit) return;
+
+		/* Do not allow allocating more than the allocation limit. */
+		this->error_thrown = true;
+		std::string msg = fmt::format("Maximum memory allocation exceeded by {} bytes when allocating {} bytes",
+			this->allocated_size + requested_size - this->allocation_limit, requested_size);
+		throw Script_FatalError(msg);
 	}
 
 	/**
-	 * Catch all validation for the allocation; did it allocate too much memory according
-	 * to the allocation limit or did the allocation at the OS level maybe fail? In those
-	 * error situations a Script_FatalError is thrown, but once that has been done further
-	 * allocations are allowed to make it possible for Squirrel to throw the error and
-	 * clean everything up.
-	 * @param requested_size The requested size that was requested to be allocated.
-	 * @param p              The pointer to the allocated object, or null if allocation failed.
+	 * Internal helper to allocate the given amount of bytes.
+	 * @param requested_size The requested size.
+	 * @return The allocated memory.
+	 * @throws Script_FatalError When memory could not be allocated.
 	 */
-	void CheckAllocation(size_t requested_size, void *p)
+	void *DoAlloc(SQUnsignedInteger requested_size)
 	{
-		if (this->allocated_size + requested_size > this->allocation_limit && !this->error_thrown) {
-			/* Do not allow allocating more than the allocation limit, except when an error is
-			 * already as then the allocation is for throwing that error in Squirrel, the
-			 * associated stack trace information and while cleaning up the AI. */
-			this->error_thrown = true;
-			std::string msg = fmt::format("Maximum memory allocation exceeded by {} bytes when allocating {} bytes",
-				this->allocated_size + requested_size - this->allocation_limit, requested_size);
-			/* Don't leak the rejected allocation. */
-			free(p);
-			throw Script_FatalError(msg);
-		}
+		try {
+			void *p = this->allocator.allocate(requested_size);
+			assert(p != nullptr);
+			this->allocated_size += requested_size;
 
-		if (p == nullptr) {
+#ifdef SCRIPT_DEBUG_ALLOCATIONS
+			assert(this->allocations.find(p) == this->allocations.end());
+			this->allocations[p] = requested_size;
+#endif
+			return p;
+		} catch (const std::bad_alloc &) {
 			/* The OS did not have enough memory to allocate the object, regardless of the
 			 * limit imposed by OpenTTD on the amount of memory that may be allocated. */
 			if (this->error_thrown) {
 				/* The allocation is called in the error handling of a memory allocation
 				 * failure, then not being able to allocate that small amount of memory
 				 * means there is no other choice than to bug out completely. */
-				MallocError(requested_size);
+				FatalError("Out of memory. Cannot allocate {} bytes", requested_size);
 			}
 
 			this->error_thrown = true;
@@ -94,21 +101,24 @@ struct ScriptAllocator {
 		}
 	}
 
+public:
+	size_t GetAllocatedSize() const { return this->allocated_size; }
+
+	void CheckLimit() const
+	{
+		if (this->allocated_size > this->allocation_limit) throw Script_FatalError("Maximum memory allocation exceeded");
+	}
+
+	void Reset()
+	{
+		assert(this->allocated_size == 0);
+		this->error_thrown = false;
+	}
+
 	void *Malloc(SQUnsignedInteger size)
 	{
-		void *p = malloc(size);
-
-		this->CheckAllocation(size, p);
-
-		this->allocated_size += size;
-
-#ifdef SCRIPT_DEBUG_ALLOCATIONS
-		assert(p != nullptr);
-		assert(this->allocations.find(p) == this->allocations.end());
-		this->allocations[p] = size;
-#endif
-
-		return p;
+		this->CheckAllocationAllowed(size);
+		return this->DoAlloc(size);
 	}
 
 	void *Realloc(void *p, SQUnsignedInteger oldsize, SQUnsignedInteger size)
@@ -121,30 +131,11 @@ struct ScriptAllocator {
 			return nullptr;
 		}
 
-#ifdef SCRIPT_DEBUG_ALLOCATIONS
-		assert(this->allocations[p] == oldsize);
-		this->allocations.erase(p);
-#endif
-		/* Can't use realloc directly because memory limit check.
-		 * If memory exception is thrown, the old pointer is expected
-		 * to be valid for engine cleanup.
-		 */
-		void *new_p = malloc(size);
+		this->CheckAllocationAllowed(size - oldsize);
 
-		this->CheckAllocation(size - oldsize, new_p);
-
-		/* Memory limit test passed, we can copy data and free old pointer. */
-		memcpy(new_p, p, std::min(oldsize, size));
-		free(p);
-
-		this->allocated_size -= oldsize;
-		this->allocated_size += size;
-
-#ifdef SCRIPT_DEBUG_ALLOCATIONS
-		assert(new_p != nullptr);
-		assert(this->allocations.find(p) == this->allocations.end());
-		this->allocations[new_p] = size;
-#endif
+		void *new_p = this->DoAlloc(size);
+		std::copy_n(static_cast<std::byte *>(p), std::min(oldsize, size), static_cast<std::byte *>(new_p));
+		this->Free(p, oldsize);
 
 		return new_p;
 	}
@@ -152,7 +143,7 @@ struct ScriptAllocator {
 	void Free(void *p, SQUnsignedInteger size)
 	{
 		if (p == nullptr) return;
-		free(p);
+		this->allocator.deallocate(reinterpret_cast<uint8_t*>(p), size);
 		this->allocated_size -= size;
 
 #ifdef SCRIPT_DEBUG_ALLOCATIONS
@@ -163,10 +154,8 @@ struct ScriptAllocator {
 
 	ScriptAllocator()
 	{
-		this->allocated_size = 0;
 		this->allocation_limit = static_cast<size_t>(_settings_game.script.script_max_memory_megabytes) << 20;
 		if (this->allocation_limit == 0) this->allocation_limit = SAFE_LIMIT; // in case the setting is somehow zero
-		this->error_thrown = false;
 	}
 
 	~ScriptAllocator()
@@ -177,31 +166,20 @@ struct ScriptAllocator {
 	}
 };
 
-/**
- * In the memory allocator for Squirrel we want to directly use malloc/realloc, so when the OS
- * does not have enough memory the game does not go into unrecoverable error mode and kill the
- * whole game, but rather let the AI die though then we need to circumvent MallocT/ReallocT.
- * For the rest of this code, the safeguards should be in place though!
- */
-#include "../safeguards.h"
-
 ScriptAllocator *_squirrel_allocator = nullptr;
 
-/* See 3rdparty/squirrel/squirrel/sqmem.cpp for the default allocator implementation, which this overrides */
-#ifndef SQUIRREL_DEFAULT_ALLOCATOR
 void *sq_vm_malloc(SQUnsignedInteger size) { return _squirrel_allocator->Malloc(size); }
 void *sq_vm_realloc(void *p, SQUnsignedInteger oldsize, SQUnsignedInteger size) { return _squirrel_allocator->Realloc(p, oldsize, size); }
 void sq_vm_free(void *p, SQUnsignedInteger size) { _squirrel_allocator->Free(p, size); }
-#endif
 
 size_t Squirrel::GetAllocatedMemory() const noexcept
 {
 	assert(this->allocator != nullptr);
-	return this->allocator->allocated_size;
+	return this->allocator->GetAllocatedSize();
 }
 
 
-void Squirrel::CompileError(HSQUIRRELVM vm, const SQChar *desc, const SQChar *source, SQInteger line, SQInteger column)
+void Squirrel::CompileError(HSQUIRRELVM vm, std::string_view desc, std::string_view source, SQInteger line, SQInteger column)
 {
 	std::string msg = fmt::format("Error {}:{}/{}: {}", source, line, column, desc);
 
@@ -216,7 +194,7 @@ void Squirrel::CompileError(HSQUIRRELVM vm, const SQChar *desc, const SQChar *so
 	}
 }
 
-void Squirrel::ErrorPrintFunc(HSQUIRRELVM vm, const std::string &s)
+void Squirrel::ErrorPrintFunc(HSQUIRRELVM vm, std::string_view s)
 {
 	/* Check if we have a custom print function */
 	SQPrintFunc *func = ((Squirrel *)sq_getforeignptr(vm))->print_func;
@@ -227,7 +205,7 @@ void Squirrel::ErrorPrintFunc(HSQUIRRELVM vm, const std::string &s)
 	}
 }
 
-void Squirrel::RunError(HSQUIRRELVM vm, const SQChar *error)
+void Squirrel::RunError(HSQUIRRELVM vm, std::string_view error)
 {
 	/* Set the print function to something that prints to stderr */
 	SQPRINTFUNCTION pf = sq_getprintfunc(vm);
@@ -251,11 +229,11 @@ void Squirrel::RunError(HSQUIRRELVM vm, const SQChar *error)
 
 SQInteger Squirrel::_RunError(HSQUIRRELVM vm)
 {
-	const SQChar *sErr = nullptr;
+	std::string_view view;
 
 	if (sq_gettop(vm) >= 1) {
-		if (SQ_SUCCEEDED(sq_getstring(vm, -1, &sErr))) {
-			Squirrel::RunError(vm, sErr);
+		if (SQ_SUCCEEDED(sq_getstring(vm, -1, view))) {
+			Squirrel::RunError(vm, view);
 			return 0;
 		}
 	}
@@ -264,7 +242,7 @@ SQInteger Squirrel::_RunError(HSQUIRRELVM vm)
 	return 0;
 }
 
-void Squirrel::PrintFunc(HSQUIRRELVM vm, const std::string &s)
+void Squirrel::PrintFunc(HSQUIRRELVM vm, std::string_view s)
 {
 	/* Check if we have a custom print function */
 	SQPrintFunc *func = ((Squirrel *)sq_getforeignptr(vm))->print_func;
@@ -275,57 +253,57 @@ void Squirrel::PrintFunc(HSQUIRRELVM vm, const std::string &s)
 	}
 }
 
-void Squirrel::AddMethod(const char *method_name, SQFUNCTION proc, uint nparam, const char *params, void *userdata, int size)
+void Squirrel::AddMethod(std::string_view method_name, SQFUNCTION proc, std::string_view params, void *userdata, int size)
 {
 	ScriptAllocatorScope alloc_scope(this);
 
-	sq_pushstring(this->vm, method_name, -1);
+	sq_pushstring(this->vm, method_name);
 
 	if (size != 0) {
 		void *ptr = sq_newuserdata(vm, size);
-		memcpy(ptr, userdata, size);
+		std::copy_n(static_cast<std::byte *>(userdata), size, static_cast<std::byte *>(ptr));
 	}
 
 	sq_newclosure(this->vm, proc, size != 0 ? 1 : 0);
-	if (nparam != 0) sq_setparamscheck(this->vm, nparam, params);
+	if (!params.empty()) sq_setparamscheck(this->vm, params.size(), params);
 	sq_setnativeclosurename(this->vm, -1, method_name);
 	sq_newslot(this->vm, -3, SQFalse);
 }
 
-void Squirrel::AddConst(const char *var_name, int value)
+void Squirrel::AddConst(std::string_view var_name, SQInteger value)
 {
 	ScriptAllocatorScope alloc_scope(this);
 
-	sq_pushstring(this->vm, var_name, -1);
+	sq_pushstring(this->vm, var_name);
 	sq_pushinteger(this->vm, value);
 	sq_newslot(this->vm, -3, SQTrue);
 }
 
-void Squirrel::AddConst(const char *var_name, bool value)
+void Squirrel::AddConst(std::string_view var_name, bool value)
 {
 	ScriptAllocatorScope alloc_scope(this);
 
-	sq_pushstring(this->vm, var_name, -1);
+	sq_pushstring(this->vm, var_name);
 	sq_pushbool(this->vm, value);
 	sq_newslot(this->vm, -3, SQTrue);
 }
 
-void Squirrel::AddClassBegin(const char *class_name)
+void Squirrel::AddClassBegin(std::string_view class_name)
 {
 	ScriptAllocatorScope alloc_scope(this);
 
 	sq_pushroottable(this->vm);
-	sq_pushstring(this->vm, class_name, -1);
+	sq_pushstring(this->vm, class_name);
 	sq_newclass(this->vm, SQFalse);
 }
 
-void Squirrel::AddClassBegin(const char *class_name, const char *parent_class)
+void Squirrel::AddClassBegin(std::string_view class_name, std::string_view parent_class)
 {
 	ScriptAllocatorScope alloc_scope(this);
 
 	sq_pushroottable(this->vm);
-	sq_pushstring(this->vm, class_name, -1);
-	sq_pushstring(this->vm, parent_class, -1);
+	sq_pushstring(this->vm, class_name);
+	sq_pushstring(this->vm, parent_class);
 	if (SQ_FAILED(sq_get(this->vm, -3))) {
 		Debug(misc, 0, "[squirrel] Failed to initialize class '{}' based on parent class '{}'", class_name, parent_class);
 		Debug(misc, 0, "[squirrel] Make sure that '{}' exists before trying to define '{}'", parent_class, class_name);
@@ -342,7 +320,7 @@ void Squirrel::AddClassEnd()
 	sq_pop(vm, 1);
 }
 
-bool Squirrel::MethodExists(HSQOBJECT instance, const char *method_name)
+bool Squirrel::MethodExists(HSQOBJECT instance, std::string_view method_name)
 {
 	assert(!this->crashed);
 	ScriptAllocatorScope alloc_scope(this);
@@ -351,7 +329,7 @@ bool Squirrel::MethodExists(HSQOBJECT instance, const char *method_name)
 	/* Go to the instance-root */
 	sq_pushobject(this->vm, instance);
 	/* Find the function-name inside the script */
-	sq_pushstring(this->vm, method_name, -1);
+	sq_pushstring(this->vm, method_name);
 	if (SQ_FAILED(sq_get(this->vm, -2))) {
 		sq_settop(this->vm, top);
 		return false;
@@ -395,7 +373,7 @@ void Squirrel::CollectGarbage()
 	sq_collectgarbage(this->vm);
 }
 
-bool Squirrel::CallMethod(HSQOBJECT instance, const char *method_name, HSQOBJECT *ret, int suspend)
+bool Squirrel::CallMethod(HSQOBJECT instance, std::string_view method_name, HSQOBJECT *ret, int suspend)
 {
 	assert(!this->crashed);
 	ScriptAllocatorScope alloc_scope(this);
@@ -410,7 +388,7 @@ bool Squirrel::CallMethod(HSQOBJECT instance, const char *method_name, HSQOBJECT
 	/* Go to the instance-root */
 	sq_pushobject(this->vm, instance);
 	/* Find the function-name inside the script */
-	sq_pushstring(this->vm, method_name, -1);
+	sq_pushstring(this->vm, method_name);
 	if (SQ_FAILED(sq_get(this->vm, -2))) {
 		Debug(misc, 0, "[squirrel] Could not find '{}' in the class", method_name);
 		sq_settop(this->vm, top);
@@ -429,16 +407,19 @@ bool Squirrel::CallMethod(HSQOBJECT instance, const char *method_name, HSQOBJECT
 	return true;
 }
 
-bool Squirrel::CallStringMethod(HSQOBJECT instance, const char *method_name, std::string *res, int suspend)
+bool Squirrel::CallStringMethod(HSQOBJECT instance, std::string_view method_name, std::string *res, int suspend)
 {
 	HSQOBJECT ret;
 	if (!this->CallMethod(instance, method_name, &ret, suspend)) return false;
-	if (ret._type != OT_STRING) return false;
-	*res = StrMakeValid(ObjectToString(&ret));
+
+	auto str = ObjectToString(&ret);
+	if (!str.has_value()) return false;
+
+	*res = StrMakeValid(*str);
 	return true;
 }
 
-bool Squirrel::CallIntegerMethod(HSQOBJECT instance, const char *method_name, int *res, int suspend)
+bool Squirrel::CallIntegerMethod(HSQOBJECT instance, std::string_view method_name, int *res, int suspend)
 {
 	HSQOBJECT ret;
 	if (!this->CallMethod(instance, method_name, &ret, suspend)) return false;
@@ -447,7 +428,7 @@ bool Squirrel::CallIntegerMethod(HSQOBJECT instance, const char *method_name, in
 	return true;
 }
 
-bool Squirrel::CallBoolMethod(HSQOBJECT instance, const char *method_name, bool *res, int suspend)
+bool Squirrel::CallBoolMethod(HSQOBJECT instance, std::string_view method_name, bool *res, int suspend)
 {
 	HSQOBJECT ret;
 	if (!this->CallMethod(instance, method_name, &ret, suspend)) return false;
@@ -466,11 +447,10 @@ bool Squirrel::CallBoolMethod(HSQOBJECT instance, const char *method_name, bool
 	sq_pushroottable(vm);
 
 	if (prepend_API_name) {
-		std::string prepended_class_name = engine->GetAPIName();
-		prepended_class_name += class_name;
-		sq_pushstring(vm, prepended_class_name, -1);
+		std::string prepended_class_name = fmt::format("{}{}", engine->GetAPIName(), class_name);
+		sq_pushstring(vm, prepended_class_name);
 	} else {
-		sq_pushstring(vm, class_name, -1);
+		sq_pushstring(vm, class_name);
 	}
 
 	if (SQ_FAILED(sq_get(vm, -2))) {
@@ -510,8 +490,25 @@ bool Squirrel::CreateClassInstance(const std::string &class_name, void *real_ins
 	return Squirrel::CreateClassInstanceVM(this->vm, class_name, real_instance, instance, nullptr);
 }
 
-Squirrel::Squirrel(const char *APIName) :
-	APIName(APIName), allocator(new ScriptAllocator())
+/* static */ SQUserPointer Squirrel::GetRealInstance(HSQUIRRELVM vm, int index, std::string_view tag)
+{
+	if (index < 0) index += sq_gettop(vm) + 1;
+	Squirrel *engine = static_cast<Squirrel *>(sq_getforeignptr(vm));
+	std::string class_name = fmt::format("{}{}", engine->GetAPIName(), tag);
+	sq_pushroottable(vm);
+	sq_pushstring(vm, class_name);
+	sq_get(vm, -2);
+	sq_push(vm, index);
+	if (sq_instanceof(vm) == SQTrue) {
+		sq_pop(vm, 3);
+		SQUserPointer ptr = nullptr;
+		if (SQ_SUCCEEDED(sq_getinstanceup(vm, index, &ptr, nullptr))) return ptr;
+	}
+	throw sq_throwerror(vm, fmt::format("parameter {} has an invalid type ; expected: '{}'", index - 1, class_name));
+}
+
+Squirrel::Squirrel(std::string_view api_name) :
+	api_name(api_name), allocator(std::make_unique<ScriptAllocator>())
 {
 	this->Initialize();
 }
@@ -539,7 +536,7 @@ void Squirrel::Initialize()
 	sq_setforeignptr(this->vm, this);
 
 	sq_pushroottable(this->vm);
-	squirrel_register_global_std(this);
+	squirrel_register_global_std(*this);
 
 	/* Set consts table as delegate of root table, so consts/enums defined via require() are accessible */
 	sq_pushconsttable(this->vm);
@@ -548,72 +545,76 @@ void Squirrel::Initialize()
 
 class SQFile {
 private:
-	FILE *file;
+	FileHandle file;
 	size_t size;
 	size_t pos;
+	std::string buffer;
+	StringConsumer consumer;
+
+	size_t ReadInternal(std::span<char> buf)
+	{
+		size_t count = buf.size();
+		if (this->pos + count > this->size) {
+			count = this->size - this->pos;
+		}
+		if (count > 0) count = fread(buf.data(), 1, count, this->file);
+		this->pos += count;
+		return count;
+	}
 
 public:
-	SQFile(FILE *file, size_t size) : file(file), size(size), pos(0) {}
+	SQFile(FileHandle file, size_t size) : file(std::move(file)), size(size), pos(0), consumer(buffer) {}
 
-	size_t Read(void *buf, size_t elemsize, size_t count)
+	StringConsumer &GetConsumer(size_t min_size = 64)
 	{
-		assert(elemsize != 0);
-		if (this->pos + (elemsize * count) > this->size) {
-			count = (this->size - this->pos) / elemsize;
+		if (this->consumer.GetBytesLeft() < min_size && this->pos < this->size) {
+			this->buffer.erase(0, this->consumer.GetBytesRead());
+
+			size_t buffer_size = this->buffer.size();
+			size_t read_size = Align(min_size - buffer_size, 4096); // read pages of 4096 bytes
+			/* TODO C++23: use std::string::resize_and_overwrite() */
+			this->buffer.resize(buffer_size + read_size);
+			auto dest = std::span(this->buffer.data(), this->buffer.size()).subspan(buffer_size);
+			buffer_size += this->ReadInternal(dest);
+			this->buffer.resize(buffer_size);
+
+			this->consumer = StringConsumer(this->buffer);
 		}
-		if (count == 0) return 0;
-		size_t ret = fread(buf, elemsize, count, this->file);
-		this->pos += ret * elemsize;
-		return ret;
+		return this->consumer;
+	}
+
+	size_t Read(void *buf, size_t max_size)
+	{
+		std::span<char> dest(reinterpret_cast<char *>(buf), max_size);
+
+		auto view = this->consumer.Read(max_size);
+		std::copy(view.data(), view.data() + view.size(), dest.data());
+		size_t result_size = view.size();
+
+		if (result_size < max_size) {
+			assert(!this->consumer.AnyBytesLeft());
+			result_size += this->ReadInternal(dest.subspan(result_size));
+		}
+
+		return result_size;
 	}
 };
 
 static char32_t _io_file_lexfeed_ASCII(SQUserPointer file)
 {
-	unsigned char c;
-	if (((SQFile *)file)->Read(&c, sizeof(c), 1) > 0) return c;
-	return 0;
+	StringConsumer &consumer = reinterpret_cast<SQFile *>(file)->GetConsumer();
+	return consumer.TryReadUint8().value_or(0); // read as unsigned, otherwise integer promotion breaks it
 }
 
 static char32_t _io_file_lexfeed_UTF8(SQUserPointer file)
 {
-	char buffer[5];
-
-	/* Read the first character, and get the length based on UTF-8 specs. If invalid, bail out. */
-	if (((SQFile *)file)->Read(buffer, sizeof(buffer[0]), 1) != 1) return 0;
-	uint len = Utf8EncodedCharLen(buffer[0]);
-	if (len == 0) return -1;
-
-	/* Read the remaining bits. */
-	if (len > 1 && ((SQFile *)file)->Read(buffer + 1, sizeof(buffer[0]), len - 1) != len - 1) return 0;
-
-	/* Convert the character, and when definitely invalid, bail out as well. */
-	char32_t c;
-	if (Utf8Decode(&c, buffer) != len) return -1;
-
-	return c;
-}
-
-static char32_t _io_file_lexfeed_UCS2_no_swap(SQUserPointer file)
-{
-	unsigned short c;
-	if (((SQFile *)file)->Read(&c, sizeof(c), 1) > 0) return (char32_t)c;
-	return 0;
-}
-
-static char32_t _io_file_lexfeed_UCS2_swap(SQUserPointer file)
-{
-	unsigned short c;
-	if (((SQFile *)file)->Read(&c, sizeof(c), 1) > 0) {
-		c = ((c >> 8) & 0x00FF)| ((c << 8) & 0xFF00);
-		return (char32_t)c;
-	}
-	return 0;
+	StringConsumer &consumer = reinterpret_cast<SQFile *>(file)->GetConsumer();
+	return consumer.AnyBytesLeft() ? consumer.ReadUtf8(-1) : 0;
 }
 
 static SQInteger _io_file_read(SQUserPointer file, SQUserPointer buf, SQInteger size)
 {
-	SQInteger ret = ((SQFile *)file)->Read(buf, 1, size);
+	SQInteger ret = reinterpret_cast<SQFile *>(file)->Read(buf, size);
 	if (ret == 0) return -1;
 	return ret;
 }
@@ -622,63 +623,47 @@ SQRESULT Squirrel::LoadFile(HSQUIRRELVM vm, const std::string &filename, SQBool
 {
 	ScriptAllocatorScope alloc_scope(this);
 
-	FILE *file;
+	std::optional<FileHandle> file = std::nullopt;
 	size_t size;
-	if (strncmp(this->GetAPIName(), "AI", 2) == 0) {
+	if (this->GetAPIName().starts_with("AI")) {
 		file = FioFOpenFile(filename, "rb", AI_DIR, &size);
-		if (file == nullptr) file = FioFOpenFile(filename, "rb", AI_LIBRARY_DIR, &size);
-	} else if (strncmp(this->GetAPIName(), "GS", 2) == 0) {
+		if (!file.has_value()) file = FioFOpenFile(filename, "rb", AI_LIBRARY_DIR, &size);
+	} else if (this->GetAPIName().starts_with("GS")) {
 		file = FioFOpenFile(filename, "rb", GAME_DIR, &size);
-		if (file == nullptr) file = FioFOpenFile(filename, "rb", GAME_LIBRARY_DIR, &size);
+		if (!file.has_value()) file = FioFOpenFile(filename, "rb", GAME_LIBRARY_DIR, &size);
 	} else {
 		NOT_REACHED();
 	}
 
-	if (file == nullptr) {
+	if (!file.has_value()) {
 		return sq_throwerror(vm, "cannot open the file");
 	}
 	unsigned short bom = 0;
 	if (size >= 2) {
-		[[maybe_unused]] size_t sr = fread(&bom, 1, sizeof(bom), file);
+		if (fread(&bom, 1, sizeof(bom), *file) != sizeof(bom)) return sq_throwerror(vm, "cannot read the file");;
 	}
 
 	SQLEXREADFUNC func;
 	switch (bom) {
 		case SQ_BYTECODE_STREAM_TAG: { // BYTECODE
-			if (fseek(file, -2, SEEK_CUR) < 0) {
-				FioFCloseFile(file);
+			if (fseek(*file, -2, SEEK_CUR) < 0) {
 				return sq_throwerror(vm, "cannot seek the file");
 			}
 
-			SQFile f(file, size);
+			SQFile f(std::move(*file), size);
 			if (SQ_SUCCEEDED(sq_readclosure(vm, _io_file_read, &f))) {
-				FioFCloseFile(file);
 				return SQ_OK;
 			}
-			FioFCloseFile(file);
 			return sq_throwerror(vm, "Couldn't read bytecode");
 		}
-		case 0xFFFE:
-			/* Either this file is encoded as big-endian and we're on a little-endian
-			 * machine, or this file is encoded as little-endian and we're on a big-endian
-			 * machine. Either way, swap the bytes of every word we read. */
-			func = _io_file_lexfeed_UCS2_swap;
-			size -= 2; // Skip BOM
-			break;
-		case 0xFEFF:
-			func = _io_file_lexfeed_UCS2_no_swap;
-			size -= 2; // Skip BOM
-			break;
 		case 0xBBEF:   // UTF-8
 		case 0xEFBB: { // UTF-8 on big-endian machine
 			/* Similarly, check the file is actually big enough to finish checking BOM */
 			if (size < 3) {
-				FioFCloseFile(file);
 				return sq_throwerror(vm, "I/O error");
 			}
 			unsigned char uc;
-			if (fread(&uc, 1, sizeof(uc), file) != sizeof(uc) || uc != 0xBF) {
-				FioFCloseFile(file);
+			if (fread(&uc, 1, sizeof(uc), *file) != sizeof(uc) || uc != 0xBF) {
 				return sq_throwerror(vm, "Unrecognized encoding");
 			}
 			func = _io_file_lexfeed_UTF8;
@@ -688,19 +673,16 @@ SQRESULT Squirrel::LoadFile(HSQUIRRELVM vm, const std::string &filename, SQBool
 		default: // ASCII
 			func = _io_file_lexfeed_ASCII;
 			/* Account for when we might not have fread'd earlier */
-			if (size >= 2 && fseek(file, -2, SEEK_CUR) < 0) {
-				FioFCloseFile(file);
+			if (size >= 2 && fseek(*file, -2, SEEK_CUR) < 0) {
 				return sq_throwerror(vm, "cannot seek the file");
 			}
 			break;
 	}
 
-	SQFile f(file, size);
+	SQFile f(std::move(*file), size);
 	if (SQ_SUCCEEDED(sq_compile(vm, func, &f, filename.c_str(), printerror))) {
-		FioFCloseFile(file);
 		return SQ_OK;
 	}
-	FioFCloseFile(file);
 	return SQ_ERROR;
 }
 
@@ -752,10 +734,8 @@ void Squirrel::Uninitialize()
 	sq_pop(this->vm, 1);
 	sq_close(this->vm);
 
-	assert(this->allocator->allocated_size == 0);
-
 	/* Reset memory allocation errors. */
-	this->allocator->error_thrown = false;
+	this->allocator->Reset();
 }
 
 void Squirrel::Reset()
